### 
###  RUNNING TIME AND COMPLEXITY
### 
###  Fredo Durand, MIT , 6.0001
###


import time
import math

def convert_to_km(m):
    return m * 1.609

def compound(invest, interest, n_months):
    total=0
    for i in range(n_months):
       total = total * interest + invest
    return total


L_N = [1]
for i in range(7):
    L_N.append(L_N[-1]*10)


# for N in L_N:
#     t = time.perf_counter()
#     km = convert_to_km(N)
#     dt = time.perf_counter()-t
#     print ("convert(", N, ") took ", dt, "seconds")


# for N in L_N:
#     t = time.perf_counter()
#     money = compound(10.0, 1.05, N)
#     dt = time.perf_counter()-t
#     print ("compound(", N, ") took ", dt, "seconds")

### run the computation n_samples times for better measurement

n_samples = 50

# for N in L_N:
#     t = time.perf_counter()
#     for i in range(n_samples):
#         km = convert_to_km(N)
#     dt = (time.perf_counter()-t)/n_samples
#     print ("convert(", N, ") took ", dt, "seconds")


# following code is longer and easier to measure (also slower) 
# so use smaller n_sample
n_samples = 5

# for N in L_N:
#     t = time.perf_counter()
#     for i in range(n_samples):  
#         money = compound(10.0, 1.05, N)
#     dt = (time.perf_counter()-t)/n_samples
#     print ("compound(", N, ") took ", dt, "seconds")


#################################
###
### SUM
##  
#################################
   
def sum_of(L):
    total = 0.0
    for elt in L:
        total = total + elt
    return total


L_N = [1]
for i in range(8):
    L_N.append(L_N[-1]*10)

# for N in L_N:
#     L = range(N)
#     t = time.perf_counter()
#     s = sum_of(L)
#     print ("sum of ", N, "elements took ", time.perf_counter()-t, "seconds")

#################################
###
### IS_IN (BINARY SEARCH)
##  
#################################

def is_in(L, x):
    for elt in L:
        if elt==x: return True
    return False

def binary_search(L, x):
    """ returns True if x is in L
        L must be sorted """
    lo = 0
    hi = len(L)
    while hi-lo > 1:
        mid = (hi+lo) // 2
        if L[mid] <= x:
            lo = mid
        else:
            hi = mid
    return L[lo] == x

L_N = [1]
for i in range(8):
    L_N.append(L_N[-1]*10)

prev_time_brute = None
prev_time_binary = None
prev_time_builtin = None

# for N in L_N:
#     L = range(N)
#     t = time.perf_counter()
#     for x in [L[0], L[len(L)//2], L[-1]]:
#         my_bool = is_in(L, x)
#     t = (time.perf_counter()-t) / 3
#     print ('for ', N, 'element, brute-force took on average', t, 's')
#     if not N == L_N[0]:
#         print ('    ', t/prev_time_brute, ' times more than for 10 times fewer elements')
#     prev_time_brute = t
    
#     t = time.perf_counter()
#     for x in [L[0], L[len(L)//2], L[-1]]:
#         my_bool = binary_search(L, x)
#     t = (time.perf_counter()-t)/3
#     print ('for ', N, 'element, binary search took on average', t, 's')
#     if not N == L_N[0]:
#         print ('    ', t/prev_time_binary, ' times more than for 10 times fewer elements')
#     prev_time_binary = t
    
#     t = time.perf_counter()
#     for x in [L[0], L[len(L)//2], L[-1]]:
#         my_bool = x in L
#     t = (time.perf_counter()-t)/3
#     print ('for ', N, 'element, builtin in took on average', t, 's')
#     if not N == L_N[0]:
#         print ('    ', t/prev_time_builtin, ' times more than for 10 times fewer elements')
#     prev_time_builtin = t
#     print

#################################
###
### POINT DIAMETER
##  
#################################

def diameter(L):
    """ assumes that L is a list of pairs of Cartesian coordinates
    compares all pairs of points and returns the largest distance """
    farthest_dist = 0
    for i in range(len(L)):
        # next loop only goes from i+1 so that pairs are not compared twice
        for j in range(i+1, len(L)):
            p1 = L[i]
            p2 = L[j]
            dist = math.sqrt( (p1[0]-p2[0])**2
                           + (p1[1]-p2[1])**2 )
            if dist > farthest_dist:
                farthest_dist = dist
    return farthest_dist

def create_list_of_2D_points(N):
    L = []
    for i in range(N):
        x = math.cos(i)*i
        y = math.sin(i)*i
        L.append((x, y))
    return L



L_N = [10, 100, 1000, 10000]

# for N in L_N:
#     L = create_list_of_2D_points(N)
#     t = time.perf_counter()
#     d = diameter(L)
#     dt = time.perf_counter() - t
#     print ("diameter of", N, "points took", dt, "seconds")

#################################
###
### BINARY NUMBERS
##  
#################################

def all_binary_numbers(N):
    def helper (prefix, N):
        if N==0:
            return [prefix]
        return helper(prefix+'0', N-1) + helper(prefix+'1', N-1)                        
    return helper('', N)

# for N in range(5):
#     print (all_binary_numbers(N))

L_N = [5, 10, 15, 20, 25]

# for N in L_N:
#     t = time.perf_counter()
#     abn = all_binary_numbers(N)
#     t = time.perf_counter()-t
#     print ('binary numbers of ', N, 'digits took ', t, 's')


#################################
###
### COUNTING OPERATIONS
##  
#################################


def is_in_counter(L, x):
    global count
    for elt in L:
        count += 1 #if == test
        if elt==x: return True
    return False

def binary_search_counter(L, x):
    """ returns True if x is in L
        L must be sorted """
    global count
    lo = 0
    hi = len(L)
    while hi-lo > 1:
        count +=1 #while test
        mid = (hi+lo) // 2
        count += 2 #mid + and /
        if L[mid] <= x:
            lo = mid
        else:
            hi = mid
        count += 1 #if test
    count += 1 #return == test
    return L[lo] == x

L_N = [1]
for i in range(6):
    L_N.append(L_N[-1]*10)

# for N in L_N:
#     L = range(N)
#     count = 0
#     # try three queries to test average time
#     for x in [L[0], L[len(L)//2], L[-1]]:
#         my_bool = is_in_counter(L, x)
#     count = count/3.0
#     print('for ', N, 'element, brute-force took on average', int(count), 'ops')
#     if not N == L_N[0]:
#         print ('    ', count/prev_time_brute, ' times more than for 10 times fewer elements')
#     prev_time_brute = count
    
#     count = 0
#     # try three queries to test average time
#     for x in [L[0], L[len(L)//2], L[-1]]:
#         my_bool = binary_search_counter(L, x)
#     count = count/3.0
#     print('for ', N, 'element, binary search took on average', int(count), 'ops')
#     if not N == L_N[0]:
#         print ('    ', count/prev_time_binary, ' times more than for 10 times fewer elements')
#     prev_time_binary = count

#     print()
    


#################################
###
### FACTORIAL PERFORMANCE IN PYTHON
###    
### 
### Complex because python upgrades integer representation when numbers get too big
###
#################################


L_N = [1]
for i in range(6):
    L_N.append(L_N[-1]*10)

def fact(N):
    f=1
    for i in range(2,N+1):
        f *= i
    return f

#     print ("convert(", N, ") took ", dt, "seconds")

# for N in L_N:
#     t = time.perf_counter()
#     f = fact(N)
#     dt = time.perf_counter()-t
#     print ("iterative fact(", N, ") took ", dt, "seconds")

